==> siggen-1.1.tgz: siggen-1.1/sgen.c <==
/* sgen.c
 * makes a raw 8 or 16 bit digital sound signal of a waveform of a particular
 * frequency and sampling rate. Output is to a file, or /dev/dsp (linux)
 * Jim Jackson     Linux Version
 */

/*
 * Copyright (C) 1997 Jim Jackson                    jj@scs.leeds.ac.uk
 *                    School of Computer Studies,
 *                    The University of Leeds,
 *                    Leeds, LS2 9JT, UK
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with this program - see the file COPYING; if not, write to 
 *  the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, 
 *  MA 02139, USA.
 */

/* History...................
 * 
 * Based on my noddy DOS version which created sample files to be played
 *   by other utils. This had lotsa faults, now fixed.
 * 
 * 15Mar97  Bug in initialising second channel to silence - it re-init first
 *          channel again. Reported by Michael Meifert <mime@dk3hg.ampr.org>
 * 10Feb97  Added -w option to write out samples as a wave file
 *   ---    for signal generation notes etc see generator.c
 * 29Dec96  Added amplitude factor -A N where N is a percentage.
 *          The sample is created to optimally fill the sample space when
 *          N is 100 (the default value). The samples generated are scaled
 *          by N/100, overly large values simply being 'clipped'.
 *          To create a trapezoid wave form, generate a triangle wave
 *          with N>100, depending on slope needed on waveform.
 * 18Dec96  Started splitting up stuff into different files so that I can 
 *          write some different front ends. All the code to create the 
 *          samples is in generator.c, and some misc. stuff in misc.c
 * --Dec96  Added noise generator using the random function - sounds ok
 *          Need to figure what to do to get pink noise - I think!
 * --Oct96  Original Linux version. Fixed faulty sample generation
 *          in DOS program - you have to generate samples over several
 *          cycles to ensure that a you can put samples end-to-end
 *          to fill a second, see comments in mksin(). Added stereo
 *          and antiphase stuff. Eventually worked out how to generate
 *          square/pulse signals accurately, and sawtooth. 
 *          Triangle still to do.
 */

/* In mono only one buffer is used to hold one secs worth of samples
 *   because we only allow integers to specify the frequency, we always
 *   have an exact number of full cycles in one sec, therefore this buffer
 *   can be treated cyclically to generate the signal for any period.
 * 
 * In stereo, two buffers hold the signal for each channel - 2nd channel is
 *   either an in phase or antiphase copy of channel 1 at moment - and they 
 *   are mixed into a large double size playing buffer.
 */

#include <stdio.h>
#include <strings.h>
#include <errno.h>
#include <ctype.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <math.h>
#include "fmtheaders.h"

#define VERSION "sgen  Ver. 1.2 (Dec 96)   Digital Signal Generator"

#define ABUF_SIZE 4096
#define DAC_FILE "/dev/dsp"
#define SAMPLERATE 22050
#define EFAIL -1

#define MAXPRM 32
#define chkword(a,b) ((n>=a)&&(strncmp(p,b,n)==0))

int aflg,fflg,vflg,dflg,wfmt;

unsigned int samplerate,freq;    /* Samples/sec and signal frequency */
unsigned int ratio;              /* used in pulse, sweep etc */
unsigned int antiphase;          /* 2nd channel in antiphase to 1st channel */
unsigned int Gain;               /* Amplification factor */
unsigned int stereo;             /* stereo mono */
unsigned int afmt;               /* format for DSP  */

unsigned int abuf_size;          /* audio buffer size */
unsigned char *abuf;             /* pointer to audio buffer */

int lbuf_flag,rbuf_flag;       
unsigned char *lbuf,*rbuf;       /* waveform bufs, left and right channels */
unsigned int lbuf_size;          /* size of 1 sec, 1 chan, waveform buffer */
unsigned char *plbuf;            /* play buffer - mixed left and right bufs */
int plbuf_size;                  /* size of playback buffer */

char *sys,*outfile;

help(e)
int e;
{  fputs(VERSION,stderr);
   fputs("\nUsage: \n 1: sgen [flags] waveform freq\n",stderr);
   fputs("      waveform is square, triangle, sawtooth, noise\n",stderr);
   fputs(" 2: sgen [flags] sin|cos freq [phase]\n",stderr);
   fputs("      sin/cos has extra phase param (def. is 0 degrees)\n",stderr);
   fputs(" 3: sgen [flags] pulse freq [Mark/Space]\n",stderr);
   fputs("      pulse has extra param Mark/Space % - def. is 10 (%)\n",stderr);
/*   fputs(" 4: sgen [flags] sweep freq [end_freq secs]\n",stderr); */
/*   fputs("      generates a swept sine wave from freq to end_freq in secs seconds\n\n",stderr); */
   fputs("Defaults: output to /dev/dsp, 22050 samples/sec, mono, 16 bit\n",stderr);
   fputs("          samples if possible, else 8 bit. \n\n",stderr);
   fputs("flags: -f,-a         force overwrite/append of/to file\n",stderr);
   fputs("       -o file       write digital sample to file ('-' is stdout)\n",stderr);
   fputs("       -w file       as '-o' but written as a wave file\n",stderr);
   fputs("       -s samples    generate with samplerate of samples/sec\n",stderr);
   fputs("       -v            be verbose.\n",stderr);
   fputs(" -8/-16 or -b 8|16   force 8 bit or 16 bit mode.\n",stderr);
   fputs("       -1,-2a        mono (def) or stereo in antiphase\n",stderr);
   fputs("       -A n          scale samples by n/100, def. n is 100\n",stderr);
   return(e);
}

/* main
 *
 */
 
main(argc,argv)
int argc;
char **argv;
{
   unsigned int v[MAXPRM],maxv,i,j,k,l,m,n,N;
   FILE *f;
   char *p,*wf,*fnm,fname[130],*omode;
   int fd,st;
   unsigned int t;
   
   if ((p=strrchr(sys=*argv++,'/'))!=NULL) sys=++p;
   argc--;
   
   outfile=NULL;
   samplerate=SAMPLERATE; afmt=AFMT_QUERY;
   antiphase=stereo=wfmt=0; Gain=100;
   dflg=vflg=aflg=fflg=0;
   
   while (argc && **argv=='-') {          /* all flags and options must come */
      n=strlen(p=(*argv++)+1); argc--;    /* before paramters                */
      if (chkword(1,"samplerate")) {
	 if (argc && isdigit(**argv)) { samplerate=atoi(*argv++); argc--; }
      }
      else if (chkword(1,"output")) {     /* e.g. string option              */
	 if (argc) { outfile=*argv++; argc--; } /* output file name          */
      }
      else if (chkword(1,"w")) {     /* e.g. string option              */
	 if (argc) { outfile=*argv++; argc--; wfmt=1; } /* output file name          */
      }
      else if (chkword(2,"16")) { afmt=AFMT_S16_LE; }
      else if (chkword(1,"bits")) {
	 i=0;
	 if (argc) { 
	    i=atoi(*argv++); argc--;
	 }
	 if (i==8) afmt=AFMT_U8;
	 else if (i==16) afmt=AFMT_S16_LE;
	 else exit(err_rpt(EINVAL,"must be '-b 8' or '-b 16'."));
      }
      else if (chkword(1,"A")) {
	 if (argc && isdigit(**argv)) { 
	    Gain=atoi(*argv++); argc--;
	 }
      }
      else if (chkword(2,"2a")) {
	 stereo=antiphase=1;
      }
      else {                              /* check for single char. flags    */
	 for (; *p; p++) {
	    if (*p=='h') exit(help(EFAIL));
	    else if (*p=='1') stereo=0;
	    else if (*p=='2') stereo=1;
	    else if (*p=='8') afmt=AFMT_U8;
	    else if (*p=='a') aflg=1;
	    else if (*p=='d') dflg=1;
	    else if (*p=='v') vflg=1;
	    else {
	       *fname='-'; *(fname+1)=*p; *(fname+2)=0;
	       exit(help(err_rpt(EINVAL,fname)));
	    }
	 }
      }
   }
   
   if (argc<2) exit(help(err_rpt(EINVAL,"Not enough parameters")));
   wf=*argv++; argc--;    /* waveform type */
   freq=atoi(*argv++); argc--;
   ratio=-1;
   if (argc) { ratio=atoi(*argv++); argc--; }

   if (argc) exit(help(err_rpt(EINVAL,"Too many parameters")));
      
   if (vflg) puts(VERSION);

   if (outfile==NULL) {           /* if no outfile then write direct to DAC */
                                  /* if no format specified then try 16 bit */
      i=(afmt==AFMT_QUERY)?AFMT_S16_LE:afmt ; 
      n=stereo;
      if ((fd=DACopen(fnm=DAC_FILE,"w",&samplerate,&i,&n))<0) {
	 if (afmt==AFMT_QUERY) {  /* if no format specified try for 8 bit.. */
	    i=AFMT_U8;
	    fd=DACopen(fnm=DAC_FILE,"w",&samplerate,&i,&n);
	 }
	 if (fd<0) exit(err_rpt(errno,"Opening DSP for output."));
      }
      afmt=i;
      if (vflg) {
	 fputs(fnm,stdout); fputs(": DAC Opened for output\n",stdout);
	 fputd(samplerate,stdout); fputs(" samples/sec\n",stdout);
	 if (i==AFMT_U8) { p="unsigned 8 bit samples, "; }
	 else if (i==AFMT_S16_LE) { p="signed 16 bit samples, little endian, "; }
	 else { p="Unknown audio format, "; }
	 fputs(p,stdout);
	 p=(stereo)?"Stereo mode (2 channels)\n":"Mono mode (1 channel)\n";
	 fputs(p,stdout);
      }
      if (i!=afmt || n!=stereo) {
	 exit(err_rpt(EINVAL,"Sound card doesn't support format requested."));
      }
   } else { 
      if (aflg && wfmt) {
         exit(err_rpt(EINVAL,"Cannot write append to a WAVE file."));
      }
      afmt=(afmt==AFMT_QUERY)?AFMT_S16_LE:afmt ; /* set 16 bit mode unless
						  * some format is forced */
      omode=(aflg)?"a":"w";
      if ((f=fopen(fnm=outfile,"r"))!=NULL) {
	 fclose(f);
	 if ( (!aflg) && (!fflg) ) exit(err_rpt(EEXIST,fnm));
      }
      if ((f=fopen(fnm,omode))==NULL) { exit(err_rpt(errno,fnm)); }
      if (vflg) { 
	 fputs(fnm,stdout);
	 if (*omode=='a') fputs(": Opened in append mode.\n",stdout);
	 else fputs(": Opened clean for writing.\n",stdout);
      }
      fd=fileno(f);
   }

   lbuf_size=samplerate;                  /* buf sizes if 8 bit samples */
   if (afmt==AFMT_S16_LE) lbuf_size<<=1;  /* double size if 16 bit      */
   else if (afmt!=AFMT_U8) {
      exit(err_rpt(EINVAL,"Only unsigned 8 and signed 16 bit, supported."));
   }
	
   lbuf_flag=rbuf_flag=0;
   lbuf=(unsigned char *)malloc(lbuf_size);
   rbuf=(unsigned char *)malloc(lbuf_size);
   plbuf=(unsigned char *)malloc(plbuf_size=lbuf_size<<stereo);
   if ((rbuf==NULL) || (lbuf==NULL) || (plbuf==NULL)) {
      exit(err_rpt(ENOMEM,"Attempting to get buffers"));
   }
   i=vflg; vflg=0;
   mknull(lbuf,lbuf_size,0,0,samplerate,0,afmt);
   mknull(rbuf,lbuf_size,0,0,samplerate,0,afmt);
   vflg=i;
   
   if (vflg) {
      fputs((stereo)?"Stereo, ":"Mono, ",stdout);
      fputs((afmt==AFMT_S16_LE)?"16":"8",stdout);
      fputs(" bit samples being generated.\n",stdout);
      fputs("Samples scaled by a factor of ",stdout);
      fputd(Gain,stdout); fputs("/100.\n",stdout);
      fputd(lbuf_size,stdout); 
      fputs(" byte buffer allocated per channel.\n",stdout);
   }

   if (freq > samplerate/2) {
      fputd(freq,stderr); 
      fputs(" Hz is more than half the sampling rate\n",stderr);
      exit(err_rpt(EINVAL,"Frequency setting too great"));
   }

   if (generate(wf,lbuf,lbuf_size,freq,Gain,samplerate,ratio,afmt)==0) {
      exit(err_rpt(ENOSYS,wf));
   }

   if (stereo) {
      if (antiphase) {
	 do_antiphase(rbuf,lbuf,samplerate,afmt);
      } else memcpy(rbuf,lbuf,lbuf_size);
      mixplaybuf(plbuf,lbuf,rbuf,samplerate,afmt);
   } else {      
      memcpy(plbuf,lbuf,plbuf_size);
   }

   if (outfile==NULL) { 
      playloop(fd,plbuf,plbuf_size);
   } else {
      if (wfmt) {
         st=writewav(fd,plbuf,plbuf_size);
      } else {
	 st=write(fd,plbuf,plbuf_size);
      }
      if (st<0) {
         err_rpt(errno,outfile);
      } else if (vflg) {
         fputs("Samples written to ",stdout);
	 fputs((wfmt)?"WAVE":"RAW",stdout);
	 fputs(" format file ",stdout); fputs(outfile,stdout);
         fputs("\n",stdout);
      }
   }
   
   free(lbuf); free(rbuf); free(plbuf);
   close(fd);
   exit(0);
}  

/*
 *  playloop(fd,bf,bfn)   bf is a sample buffer of size bfn bytes.
 *      playloop plays this continuously to file descr fd
 */

playloop(fd,bf,bfn)
int fd;
unsigned char *bf;
int bfn;
{
   int i,bn;
   unsigned char *p;
   
   bn=bfn/10;
   for (;;) {
      for ( p=bf,i=0; i< bfn; i+=bn) { 
	 if (write(fd,p+i,bn) < 0) return(0);
	 /* check here if need to supend output or handle keypresses etc*/
      }
      if ((i!=bfn) && (write(fd,p+i-bn,bfn-i+bn)<0)) return(0);
   }
}

/*
 * DACopen(char *fnm, char *mode, int *samples, int *fmt, int *stereo)
 * 
 *   open dspfile for read "r" or write "w", and set samples per sec
 *    as sampling rate - note we get pointer for samples so we can 
 *    return the actual samplerate set. Stereo is non zero for stereo,
 *    and fmt is the format (8 or 16 bit)
 *    and again these are pointers, so actually settings can be returned.
 *   return file descriptor or -1 on error.
 */

DACopen(fnm,mode,samples,fmt,stereo)
char *fnm;
char *mode;
int *samples, *stereo, *fmt;
{
   int fd;
   int m;
   
   if (*mode=='r') m=O_RDONLY;
   else if (*mode=='w') m=O_WRONLY;
   else {
      errno=EINVAL; return(-1);
   }
   if ((fd = open (fnm,m,0)) >= 0) {  /* params must be set in this order */
      if (ioctl(fd, SNDCTL_DSP_SETFMT, fmt)>=0) { 
	 if (ioctl(fd, SNDCTL_DSP_STEREO, stereo)>=0) { 
	    if (ioctl(fd, SNDCTL_DSP_SPEED, samples)>=0) { 
	       return(fd);
	    }
	 }
      }
   }
   return(-1);
}

/* writewav(fd,buf,N)  write N bytes of sample buf to fd with an 
 *               appropriate wave format header.
 */

writewav(fd,buf,N)
int fd;
char *buf;
int N;

{
   WaveHeader wh;

   wh.main_chunk = RIFF;
   wh.length     = N + sizeof(WaveHeader) - 8; 
   wh.chunk_type = WAVE;
   wh.sub_chunk  = FMT;
   wh.sc_len     = 16;
   wh.format     = PCM_CODE;
   wh.modus      = (stereo) ? 2 : 1;
   wh.sample_fq  = samplerate;
   wh.byte_p_spl = (afmt==AFMT_U8) ? 1 : 2;
   wh.byte_p_sec = samplerate * wh.modus * wh.byte_p_spl;
   wh.bit_p_spl  = (afmt==AFMT_U8) ? 8 : 16;
   wh.data_chunk = DATA;
   wh.data_length= N;
   if (write(fd, &wh, sizeof(WaveHeader))<0) return(-1);
   return(write(fd, buf, N));
}

==> siggen-2.1.tgz: siggen-2.1/sgen.c <==
/* sgen.c
 * makes a raw 8 or 16 bit digital sound signal of a waveform of a particular
 * frequency and sampling rate. Output is to a file, or /dev/dsp (linux)
 * Jim Jackson     Linux Version
 */

/*
 * Copyright (C) 1997 Jim Jackson                    jj@scs.leeds.ac.uk
 *                    School of Computer Studies,
 *                    The University of Leeds,
 *                    Leeds, LS2 9JT, UK
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with this program - see the file COPYING; if not, write to 
 *  the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, 
 *  MA 02139, USA.
 */

/* History...................
 * 
 * Based on my noddy DOS version which created sample files to be played
 *   by other utils. This had lotsa faults, now fixed.
 * 
 * 04Aug97  Added -t option to specify time to play 
 * 18Mar97  -f flag was missing - added it back.
 *          Added default sine waveform if none specified on line.
 * 15Mar97  Bug in initialising second channel to silence - it re-init first
 *          channel again. Reported by Michael Meifert <mime@dk3hg.ampr.org>
 * 10Feb97  Added -w option to write out samples as a wave file
 *   ---    for signal generation notes etc see generator.c
 * 29Dec96  Added amplitude factor -A N where N is a percentage.
 *          The sample is created to optimally fill the sample space when
 *          N is 100 (the default value). The samples generated are scaled
 *          by N/100, overly large values simply being 'clipped'.
 *          To create a trapezoid wave form, generate a triangle wave
 *          with N>100, depending on slope needed on waveform.
 * 18Dec96  Started splitting up stuff into different files so that I can 
 *          write some different front ends. All the code to create the 
 *          samples is in generator.c, and some misc. stuff in misc.c
 * --Dec96  Added noise generator using the random function - sounds ok
 *          Need to figure what to do to get pink noise - I think!
 * --Oct96  Original Linux version. Fixed faulty sample generation
 *          in DOS program - you have to generate samples over several
 *          cycles to ensure that a you can put samples end-to-end
 *          to fill a second, see comments in mksin(). Added stereo
 *          and antiphase stuff. Eventually worked out how to generate
 *          square/pulse signals accurately, and sawtooth. 
 *          Triangle still to do.
 */

/* In mono only one buffer is used to hold one secs worth of samples
 *   because we only allow integers to specify the frequency, we always
 *   have an exact number of full cycles in one sec, therefore this buffer
 *   can be treated cyclically to generate the signal for any period.
 * 
 * In stereo, two buffers hold the signal for each channel - 2nd channel is
 *   either an in phase or antiphase copy of channel 1 at moment - and they 
 *   are mixed into a large double size playing buffer.
 */

#include <stdio.h>
#include <strings.h>
#include <errno.h>
#include <ctype.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <math.h>
#include <sys/soundcard.h>

#define VERSION "sgen  Ver. 2.0 (Aug 97)   Digital Signal Generator"

#define DAC_FILE "/dev/dsp"
#define SAMPLERATE 22050
#define EFAIL -1

#define MAXPRM 32
#define chkword(a,b) ((n>=a)&&(strncmp(p,b,n)==0))

int aflg,fflg,vflg,dflg,wfmt;

unsigned int samplerate,freq;    /* Samples/sec and signal frequency */
unsigned int ratio;              /* used in pulse, sweep etc */
unsigned int antiphase;          /* 2nd channel in antiphase to 1st channel */
unsigned int Gain;               /* Amplification factor */
unsigned int stereo;             /* stereo mono */
unsigned int afmt;               /* format for DSP  */
unsigned int frag_size;          /* size of DSP buffer fragments */

int lbuf_flag,rbuf_flag;       
unsigned char *lbuf,*rbuf;       /* waveform bufs, left and right channels */
unsigned int lbuf_size;          /* size of 1 sec, 1 chan, waveform buffer */
unsigned char *plbuf;            /* play buffer - mixed left and right bufs */
unsigned char *silbuf;           /* silence buffer - used for timed output */
int plbuf_size;                  /* size of playback buffer */
long playtime,playsamples;       /* number of millisecs/samples to play
                                    0 means play for ever                  */
char *sys,*outfile;

help(e)
int e;
{  fputs(VERSION,stderr);
   fputs("\nUsage: \n 1: sgen [flags] waveform freq\n",stderr);
   fputs("      waveform is square, triangle, sawtooth, noise\n",stderr);
   fputs(" 2: sgen [flags] sin|cos freq [phase]\n",stderr);
   fputs("      sin/cos has extra phase param (def. is 0 degrees)\n",stderr);
   fputs(" 3: sgen [flags] pulse freq [Mark/Space]\n",stderr);
   fputs("      pulse has extra param Mark/Space % - def. is 10 (%)\n",stderr);
   fputs("Defaults: output continuously to /dev/dsp, 22050 samples/sec, mono,\n",stderr);
   fputs("          16 bit samples if possible, else 8 bit. \n\n",stderr);
   fputs("flags: -f,-a         force overwrite/append of/to file\n",stderr);
   fputs("       -o file       write digital sample to file ('-' is stdout)\n",stderr);
   fputs("       -w file       as '-o' but written as a wave file\n",stderr);
   fputs("       -s samples    generate with samplerate of samples/sec\n",stderr);
   fputs("       -v            be verbose.\n",stderr);
   fputs(" -8/-16 or -b 8|16   force 8 bit or 16 bit mode.\n",stderr);
   fputs("       -1,-2a        mono (def) or stereo in antiphase\n",stderr);
   fputs("       -A n          scale samples by n/100, def. n is 100\n",stderr);
   fputs("       -t N|Nm       play for N secs or Nm millisecs\n",stderr);
   return(e);
}

/* main
 *
 */
 
main(argc,argv)
int argc;
char **argv;
{
   unsigned int v[MAXPRM],maxv,i,j,k,l,m,n,N;
   FILE *f;
   char *p,*wf,*fnm,fname[130],*omode;
   int fd,st;
   unsigned int t;
   
   if ((p=strrchr(sys=*argv++,'/'))!=NULL) sys=++p;
   argc--;
   
   outfile=NULL;
   samplerate=SAMPLERATE; afmt=AFMT_QUERY;
   antiphase=stereo=wfmt=0; Gain=100;
   playtime=playsamples=dflg=vflg=aflg=fflg=0;
   
   while (argc && **argv=='-') {          /* all flags and options must come */
      n=strlen(p=(*argv++)+1); argc--;    /* before paramters                */
      if (chkword(1,"samplerate")) {
	 if (argc && isdigit(**argv)) { samplerate=atoi(*argv++); argc--; }
      }
      else if (chkword(1,"output")) {     /* e.g. string option              */
	 if (argc) { outfile=*argv++; argc--; } /* output file name          */
      }
      else if (chkword(1,"w")) {     /* e.g. string option              */
	 if (argc) { outfile=*argv++; argc--; wfmt=1; } /* output file name          */
      }
      else if (chkword(2,"16")) { afmt=AFMT_S16_LE; }
      else if (chkword(1,"bits")) {
	 i=0;
	 if (argc) { 
	    i=atoi(*argv++); argc--;
	 }
	 if (i==8) afmt=AFMT_U8;
	 else if (i==16) afmt=AFMT_S16_LE;
	 else exit(err_rpt(EINVAL,"must be '-b 8' or '-b 16'."));
      }
      else if (chkword(1,"time")) {
	 i=0;
	 if (argc) { 
	    i=strtol(*argv++,&p,10); argc--;
	    if (*p=='s' || *p==0)
	      i*=1000;        /* time specified in secs - *1000 for ms */
	    else if (*p!='m')
	      exit(err_rpt(EINVAL,"Invalid time specification."));
	 }
	 playtime=i;
      }
      else if (chkword(1,"A")) {
	 if (argc && isdigit(**argv)) { 
	    Gain=atoi(*argv++); argc--;
	 }
      }
      else if (chkword(2,"2a")) {
	 stereo=antiphase=1;
      }
      else {                              /* check for single char. flags    */
	 for (; *p; p++) {
	    if (*p=='h') exit(help(EFAIL));
	    else if (*p=='1') stereo=0;
	    else if (*p=='2') stereo=1;
	    else if (*p=='8') afmt=AFMT_U8;
	    else if (*p=='a') aflg=1;
	    else if (*p=='f') fflg=1;
	    else if (*p=='d') dflg=1;
	    else if (*p=='v') vflg++;
	    else {
	       *fname='-'; *(fname+1)=*p; *(fname+2)=0;
	       exit(help(err_rpt(EINVAL,fname)));
	    }
	 }
      }
   }
   
   if (argc==0) exit(help(err_rpt(EINVAL,"Wrong number of parameters.")));
   wf="sine";
   if (!isdigit(**argv)) {
      wf=*argv++; argc--;    /* waveform type */
   }
   if (argc) {
      freq=atoi(*argv++); argc--;
      ratio=-1;
      if (argc) { ratio=atoi(*argv++); argc--; }
   }
   if (argc) exit(help(err_rpt(EINVAL,"Wrong number of parameters.")));
      
   if (vflg) puts(VERSION);

   if (outfile==NULL) {           /* if no outfile then write direct to DAC */
                                  /* if no format specified then try 16 bit */
      i=(afmt==AFMT_QUERY)?AFMT_S16_LE:afmt ; 
      n=stereo;
      if ((fd=DACopen(fnm=DAC_FILE,"w",&samplerate,&i,&n))<0) {
	 if (afmt==AFMT_QUERY) {  /* if no format specified try for 8 bit.. */
	    i=AFMT_U8;
	    fd=DACopen(fnm=DAC_FILE,"w",&samplerate,&i,&n);
	 }
	 if (fd<0) exit(err_rpt(errno,"Opening DSP for output."));
      }
      if ((frag_size=getfragsize(fd))<0)
	exit(err_rpt(errno,"Problem getting DSP Buffer size."));

      afmt=i;
      if (vflg) {
	 printf("%s : DAC Opened for output\n",fnm);
	 printf("%d %s, %s, samples/sec.\n",samplerate,
                (stereo)?"stereo":"mono",
		(i==AFMT_U8)?"unsigned 8 bit":
		((i==AFMT_S16_LE)?"signed 16 bit, little endian":
		 "Unknown audio format"));
         printf("%d bytes per DSP buffer.\n",frag_size);
      }
      if (i!=afmt || n!=stereo) {
	 exit(err_rpt(EINVAL,"Sound card doesn't support format requested."));
      }
   } else {   /*  outfile!=NULL  so writing to a file  */
      if (aflg && wfmt) {
         exit(err_rpt(EINVAL,"Cannot write append to a WAVE file."));
      }
      afmt=(afmt==AFMT_QUERY)?AFMT_S16_LE:afmt ; /* set 16 bit mode unless
						  * some format is forced */
      omode=(aflg)?"a":"w";
      if ((f=fopen(fnm=outfile,"r"))!=NULL) {
	 fclose(f);
	 if ( (!aflg) && (!fflg) ) exit(err_rpt(EEXIST,fnm));
      }
      if ((f=fopen(fnm,omode))==NULL) { exit(err_rpt(errno,fnm)); }
      if (vflg) { 
	 fputs(fnm,stdout);
	 if (*omode=='a') fputs(": Opened in append mode.\n",stdout);
	 else fputs(": Opened clean for writing.\n",stdout);
      }
      fd=fileno(f);
   }
                                          /* calc number of samples to play */
   if (playtime) playsamples=(playtime*samplerate+500)/1000;
   
   lbuf_size=samplerate;                  /* buf sizes if 8 bit samples */
   if (afmt==AFMT_S16_LE) lbuf_size<<=1;  /* double size if 16 bit      */
   else if (afmt!=AFMT_U8) {
      exit(err_rpt(EINVAL,"Only unsigned 8 and signed 16 bit, supported."));
   }
	
   lbuf_flag=rbuf_flag=0;
   lbuf=(unsigned char *)malloc(lbuf_size);
   rbuf=(unsigned char *)malloc(lbuf_size);
   plbuf=(unsigned char *)malloc(plbuf_size=lbuf_size<<stereo);
   if ((rbuf==NULL) || (lbuf==NULL) || (plbuf==NULL)) {
      exit(err_rpt(ENOMEM,"Attempting to get buffers"));
   }

   i=vflg; vflg=0;
   if (playtime) {
      silbuf=(unsigned char *)malloc(plbuf_size);
      if (silbuf==NULL) exit(err_rpt(ENOMEM,"Attempting to get buffers"));
      mknull(silbuf,plbuf_size,0,0,samplerate,0,afmt);
   }
   mknull(lbuf,lbuf_size,0,0,samplerate,0,afmt);
   mknull(rbuf,lbuf_size,0,0,samplerate,0,afmt);
   vflg=i;
   
   if (vflg) {
      printf("%s bit %s samples being generated.\n",
	     (afmt==AFMT_S16_LE)?"16":"8",(stereo)?"Stereo":"Mono");
      if (playtime) {
	 printf("Playing for %d millisecs, %d samples.\n",
		playtime,playsamples);
      }
      printf("Samples scaled by a factor of %d/100.\n",Gain);
      printf("%d byte buffer allocated per channel.\n",lbuf_size);
   }

   if (freq > samplerate/2) {
      fprintf(stderr,"%d Hz is more than half the sampling rate\n",freq);
      exit(err_rpt(EINVAL,"Frequency setting too great"));
   }

   if (generate(wf,lbuf,lbuf_size,freq,Gain,samplerate,ratio,afmt)==0) {
      exit(err_rpt(ENOSYS,wf));
   }

   if (stereo) {
      if (antiphase) {
	 do_antiphase(rbuf,lbuf,samplerate,afmt);
      } else memcpy(rbuf,lbuf,lbuf_size);
      mixplaybuf(plbuf,lbuf,rbuf,samplerate,afmt);
   } else {      
      memcpy(plbuf,lbuf,plbuf_size);
   }

   if (outfile==NULL) { 
      if (playtime) {
	 n=(playsamples<<stereo)<<(afmt==AFMT_S16_LE);
	 st=writecyclic(fd,plbuf,plbuf_size,n);
	 writecyclic(fd,silbuf,plbuf_size,frag_size-(n%frag_size));
      } else {
         playloop(fd,plbuf,plbuf_size);
      }
   } else {
      if (playtime==0) playsamples=samplerate;   /* 1 secs worth of samples */
      n=(playsamples<<stereo)<<(afmt==AFMT_S16_LE);
      if (wfmt) {
         st=writewavhdr(fd,playsamples,samplerate,afmt,stereo);
      }
      st=writecyclic(fd,plbuf,plbuf_size,n);
      if (st<0) {
         err_rpt(errno,outfile);
      } else if (vflg) {
         printf("%d samples written to %s format file %s\n",
	        playsamples<<stereo,stdout,(wfmt)?"WAVE":"RAW",outfile);
      }
   }
   
   free(lbuf); free(rbuf); free(plbuf); free(silbuf);
   close(fd);
   exit(0);
}  


/*
 *  playloop(fd,bf,bfn)   bf is a sample buffer of size bfn bytes.
 *    
 */

playloop(fd,bf,bfn)
int fd;
unsigned char *bf;
int bfn;
{
   int i,bn;
   unsigned char *p;
   
   bn=bfn/10;
   for (;;) {
      for ( p=bf,i=0; i< bfn; i+=bn) { 
	 if (write(fd,p+i,bn) < 0) return(0);
	 /* check here if need to supend output or handle keypresses etc*/
      }
      if ((i!=bfn) && (write(fd,p+i-bn,bfn-i+bn)<0)) return(0);
   }
}
